package Function;

use Exporter;
@ISA = (Exporter);
@EXPORT = qq(gamma, lambert, legendre);

use strict;
use warnings;

# --- Other mathematical functions ---

sub beta{ 
    my $binput1 = shift;
    my $binput2 = shift;
    
    my $binput12 = $binput1+$binput2;
    my $bvalue = ( &gamma($binput1)*&gamma($binput2) )/&gamma($binput12);
    
    return $bvalue;
}


sub gamma{ #   Hasting's approximation
    my $ginput = shift;
    
    if ($ginput<=0||$ginput>35){
        return 0;
    }

    $ginput-=1;
    my $gfact = 1;
    my @gcoef = ( 1.0, -0.577191652, 0.988205891, -0.897056937, 0.918206857,-0.756704078, 0.482199394, -0.193527818, 0.035868343 );
    
    while ($ginput > 1.0){
        $gfact *= $ginput;
        $ginput--;
    }
    
    if ($ginput==1){return $gfact;}
    
    my $gvalue = $gcoef[0]+$gcoef[1]*$ginput+$gcoef[2]*$ginput**2+$gcoef[3]*$ginput**3+$gcoef[4]*$ginput**4+$gcoef[5]*$ginput**5+$gcoef[6]*$ginput**6+$gcoef[7]*$ginput**7+$gcoef[8]*$ginput**8;
    
    $gfact*=$gvalue;
    return $gfact;
}


sub Jn{
    my $n = shift;
    my $x = shift;
    my $jn = 0;
    
    for (0..35){
        my $k = $_;
        my $s = $k+$n+1;
        last if $s > 35;
        $jn += ( (-1)**$k / (&fact($k)*&gamma($s)) ) * ( ($x/2)**($n+$k*2) );
    }
    
    return $jn;
}


sub In{
    my $n = shift;
    my $x = shift;
    my $in = 0;

    for (0..35){
        my $k = $_;
        my $s = $k+$n+1;
        last if $s > 35;
        $in += ($x/2)**(2*$k) / ( &gamma($s) * &fact($k) );
    }

    $in *= ($x/2)**$n;

    return $in;
}


sub lambert{    # Corless recurrence formula
    my $winput = shift;
    my $low = -1*1/exp(1);
    if ($winput<$low){return 0;}

    my $wc = 0;
    my $limit = 1E-12;
    
    for (my $i=0; $i<100; $i++){
        my $ewc = exp($wc);
        
        if ( $limit < abs( ($winput-($wc*$ewc))/(($wc+1)*$ewc) ) ){   # Chapeau-Belandeau & Monir
            $wc = $wc-                                        # Corless
                ($wc*$ewc-$winput)
                    /
                (   ( $ewc*($wc+1) ) -
                    ( ($wc+2)*($wc*$ewc-$winput)/(2*$wc+2) )   );
        }else{last;}
    }
    return $wc;
}

sub legendre{ # recurrence formula
    my $x = shift;
    my $n = shift;
    
    return 1 if $n==0;
    return $x if $n==1;
    
    my $pn;
    
    my $pn_1 = $x;
    my $pn_2 = 1;
    
    my $tn = 2;
    
    for (0..$n-$tn){
        $pn = ( (2*$tn-1)/$tn * $x * $pn_1 - ($tn-1)/$tn*$pn_2 );
        
        $pn_2 = $pn_1;
        $pn_1 = $pn;
        $tn++;
    }
    
    return $pn;
}

sub hermite{ # recurrence formula
    my $x = shift;
    my $n = shift;
    
    return 1 if $n==0;
    return 2*$x if $n==1;
    
    my $hn;
    
    my $hn_1 = 2*$x;
    my $hn_2 = 1;
    
    my $tn = 2;
    
    for (0..$n-$tn){
        $hn = ( 2* $x * $hn_1 - 2*($tn-1)*$hn_2 );
        
        $hn_2 = $hn_1;
        $hn_1 = $hn;
        $tn++;
    }
    
    return $hn;
}

sub laguerre{ # recurrence formula
    my $x = shift;
    my $n = shift;
    
    return 1 if $n==0;
    return 1-$x if $n==1;
    
    my $ln;
    
    my $ln_1 = 1-$x;
    my $ln_2 = 1;
    
    my $tn = 2;
    
    for (0..$n-$tn){
        $ln = ( (2*$tn-1-$x) * $ln_1 - ($tn-1)*$ln_2 ) / $tn;
        
        $ln_2 = $ln_1;
        $ln_1 = $ln;
        $tn++;
    }
    
    return $ln;
}


sub nPr{ 
    my $n = shift;
    my $r = shift;
    my $sub = $n-$r;
    $n++; $r++; $sub++;
    
    my $npr = &gamma($n)/&gamma($sub);
    
    return $npr;
}

sub nCr{ 
    my $n = shift;
    my $r = shift;
    my $sub = $n-$r;
    $n++; $r++; $sub++;
     
    my $ncr = &gamma($n)/( &gamma($sub)*&gamma($r) );
    
    return $ncr;
}

sub fact{
    my $ininum = shift;
    my $fact = 1;
    for (1..$ininum){
        $fact*=$ininum;
        $ininum--;
    }
    return $fact;
}

return 1;